import logging


def etl_collection(mongo_db, collection_name, fn):
    documents = mongo_db[collection_name].find()

    logging.info(
        f"Processing {documents.count()} documents in collection "
        f"'{collection_name}'")

    for document in documents:
        try:
            fn(document)
        except Exception:
            logging.exception(
                "Error processing document in collection "
                f"'{collection_name}': {document['_id']}")



import datetime
from datetime import timezone
import os
import unittest
import urllib

from bson.objectid import ObjectId
import pymongo
import sqlalchemy
import sqlalchemy_utils

from ustack_etl.etl import ETL, CollectionETL


class DatabaseTestCase(unittest.TestCase):
    def setUp(self):
        super().setUp()

        self.mongo_client = pymongo.MongoClient(os.environ["MONGODB_URI"])
        self.mongo_db = self.mongo_client.get_database()
        self.addCleanup(lambda: self.mongo_client.drop_database(self.mongo_db))

        sql_uri = os.environ["SQL_URI"]
        parsed_sql_uri = urllib.parse.urlparse(sql_uri)

        netloc = f"testreader@{parsed_sql_uri.hostname}"
        if parsed_sql_uri.port is not None:
            netloc += f":{parsed_sql_uri.port}"

        reader_sql_uri = urllib.parse.urlunparse((
            parsed_sql_uri.scheme, netloc, parsed_sql_uri.path,
            parsed_sql_uri.params,
            parsed_sql_uri.query, parsed_sql_uri.fragment))

        sqlalchemy_utils.create_database(sql_uri)
        self.sql_engine = sqlalchemy.create_engine(sql_uri)
        self.addCleanup(lambda: sqlalchemy_utils.drop_database(sql_uri))

        with self.sql_engine.begin() as sql_conn:
            sql_conn.execute("create role testreader login")
            sql_conn.execute("create role readers")
            sql_conn.execute("grant readers to testreader")

        self.addCleanup(self._cleanup_testreader)

        self.reader_sql_engine = sqlalchemy.create_engine(reader_sql_uri)
        self.addCleanup(self.reader_sql_engine.dispose)

    def _cleanup_testreader(self):
        with self.sql_engine.begin() as sql_conn:
            sql_conn.execute("drop owned by testreader")
            sql_conn.execute("drop role testreader")
            sql_conn.execute("drop owned by readers")
            sql_conn.execute("drop role readers")

    def object_id_from_datetime(self, *args):
        return ObjectId.from_datetime(
            datetime.datetime(*args, tzinfo=timezone.utc))


class MockConsumer():
    def __init__(self, mapping_name):
        self.mapping_name = mapping_name

        self.consumes = self.mapping_name
        self.mongo_collection_name = "mock_consumer"

    def __call__(self):
        class _MockConsumerInstance(CollectionETL):
            sql_metadata = sqlalchemy.MetaData()
            mongo_collection_name = self.mongo_collection_name

            def __setattr__(inst, name, value):
                if name == self.mapping_name:
                    self.captured_state = value

                super().__setattr__(name, value)

            def process_document(inst, sql_conn, document):
                pass

        return _MockConsumerInstance()


class MockProducer():
    def __init__(self, mapping_name, mapping_value):
        self.mapping_name = mapping_name

        self.consumes = "__none__"
        self.produces = (self.mapping_name, lambda: mapping_value)
        self.mongo_collection_name = "mock_producer"

    def __call__(self):
        class _MockProducerInstance(CollectionETL):
            sql_metadata = sqlalchemy.MetaData()
            mongo_collection_name = self.mongo_collection_name

            def process_document(inst, sql_conn, document):
                pass

        return _MockProducerInstance()


class ETLTestCase(DatabaseTestCase):
    def perform_etl(self):
        etl = ETL(self.mongo_db, self.sql_engine)

        self.add_collections(etl)

        etl.perform_etl()

    def add_collections(self, etl):
        pass



from collections import defaultdict
import logging

from ustack_etl import mongo
from ustack_etl import sql


class _StateDependency():
    def __init__(self, mongo_db, sql_engine):
        self.consumers = []
        self.producer = None
        self.mongo_db = mongo_db
        self.sql_engine = sql_engine

    def set_producer(self, state_constructor, producer):
        self.state_constructor = state_constructor
        self.producer = producer

    def add_consumer(self, consumer):
        self.consumers.append(consumer)

    def _perform_single_etl(self, attribute_name, state, collection_etl_class):
        logging.info(
            "Starting ETL process for collection "
            f"''{collection_etl_class.mongo_collection_name}'")

        collection_etl = collection_etl_class()

        setattr(collection_etl, attribute_name, state)

        for table in collection_etl.sql_metadata.tables.values():
            sql.create_table(self.sql_engine, table)

        def process_document(document):
            with self.sql_engine.begin() as sql_conn:
                collection_etl.process_document(sql_conn, document)

        mongo.etl_collection(
            self.mongo_db, collection_etl.mongo_collection_name,
            process_document)

        logging.info(
            "Completed ETL process for collection "
            f"'{collection_etl_class.mongo_collection_name}'")

    def perform_etl(self, attribute_name):
        state = None

        if self.producer is not None:
            state = self.state_constructor()

            self._perform_single_etl(attribute_name, state, self.producer)

        for consumer in self.consumers:
            self._perform_single_etl(attribute_name, state, consumer)


class ETL():
    def __init__(self, mongo_db, sql_engine):
        self.mongo_db = mongo_db
        self.sql_engine = sql_engine
        self.collection_dependencies = defaultdict(
            lambda: _StateDependency(self.mongo_db, self.sql_engine))

    def add_collection(self, collection_etl_class):
        if hasattr(collection_etl_class, "produces"):
            produces_attribute, state_constructor = \
                collection_etl_class.produces

            if collection_etl_class.consumes != "__none__":
                raise ValueError(
                    f"Invalid CollectionETL: {collection_etl_class} cannot "
                    f"produce {produces_attribute} as a consumer of "
                    f"{collection_etl_class.consumes}")

            self.collection_dependencies[produces_attribute].set_producer(
                state_constructor, collection_etl_class)
        else:
            self.collection_dependencies[
                collection_etl_class.consumes].add_consumer(
                    collection_etl_class)

    def perform_etl(self):
        for attribute_name, dependency in self.collection_dependencies.items():
            dependency.perform_etl(attribute_name)


class CollectionETL():
    consumes = "__none__"



def create_table(sql_engine, table):
    with sql_engine.begin() as sql_conn:
        table.drop(sql_conn, checkfirst=True)
        table.create(sql_conn)
        sql_conn.execute(f"grant select on \"{table.name}\" to readers")


def sanitize_text(text):
    return text.replace("\x00", "")



try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup


install_requires = []

setup(name="ustack-etl",
      version="0.1.0",
      description="Framework for writing ETL processes from MongoDB to SQL",
      url="https://github.com/ustudio/ustack-etl",
      packages=["ustack_etl"],
      install_requires=install_requires)
